/* QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* coddpscan.c
 *
 * ER, Sun Dec  9 16:30:34 CST 2001 [St. Louis]
 * 
 * dynamic programming (viterbi and forward) with the CODmodel
 *
 * calculates:
 *                       P(seqX,seqY \pi^* | CODmodel)  [viterbi algorithm; \pi^* = best path ]
 *              \sum_\pi P(seqX,seqY \pi   | CODmodel)  [forward algorithm ]
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

static double bestC33diagscan(int win, int i, int imod, int l, 
			      SQINFO sqinfoX, int *seqX, SQINFO sqinfoY, int *seqY, 
			      struct coddp_s    *dp, 
			      struct codmodel_s *cod);
static double bestCediagscan(int win, int i, int imod, int l, 
			     SQINFO sqinfoX, int *seqX, SQINFO sqinfoY, int *seqY, 
			     struct coddpscan_s *dp, 
			     struct codmodel_s  *cod, 
			     struct nullmodel_s *null, 
			     struct ali_s       *ali, 
			     int logodds);


/* Function: AllocDpDiagScanCOD()
 * Date:     ER, Mon Dec 10 10:10:59 CST 2001 [St. Louis]
 *
 * Purpose:  Allocates memory for the dp matrices of the COD model - scanning diagonal DP
 *
 * Returns:  coddp are allocated
 */
struct coddpscan_s *
AllocDpDiagScanCOD(int L)
{
  struct coddpscan_s *coddpscan;       /* structure with dp matrices   */

  coddpscan = (struct coddpscan_s *) MallocOrDie (sizeof(struct coddpscan_s));

  coddpscan->cob = AllocDpDiagOTH(L);
  coddpscan->coj = AllocDpDiagOTH(L);
  coddpscan->coe = AllocDpDiagOTH(L);

  coddpscan->cod = AllocDpDiagCOD(L);
  
  return coddpscan;
}

/* Function: FreeDpScanCOD()
 * Date:     ER,  Mon Dec 10 10:25:26 CST 2001[St. Louis]
 *
 * Purpose:  frees memory for the dp matrices of the COD model
 *
 * Returns:  coddp are freed
 */
void
FreeDpScanCOD(struct coddpscan_s *coddpscan)
{
  FreeDpDiagOTH(coddpscan->cob);
  FreeDpDiagOTH(coddpscan->coj);
  FreeDpDiagOTH(coddpscan->coe);

  FreeDpCOD(coddpscan->cod);

  free(coddpscan);
}

/* Function: ViterbiOTHDiagScan() 
 * Date:     ER,  Sun Dec  9 16:31:47 CST 2001[St. Louis] 
 * 
 * Purpose:  Calculates, i \in [start, start+L-1] --- l \in [1, i+1]
 *           mt[i][l] = P(X,Y, begin = start+i-l+1, end = start+i, best align | OTH model) 
 * 
 * Args:     ofp          -- output file 
 *           seqX, seqY   -- equal length sequences, ACGT and gaps (-)  
 *           start        -- start position 
 *           L            -- length of sX,sY  
 *           oth          -- oth_model structure 
 *           dp           -- matrices for dynamic programming
 *           ali          -- array with alignment
 *           traceback    -- if true due traceback
 *           alignment    -- if true print alignment
 * 
 * Returns:  void.
 */
double
ViterbiCODDiagScan(FILE *ofp, SQINFO sqinfoX, int *seqX, SQINFO sqinfoY, int *seqY, 
		   int win, int stabs, int stmod, int l, int lmax, struct codmodel_s *cod, 
		   struct coddpscan_s *dp,
		   struct nullmodel_s *null, struct ali_s *ali, int alignment, int logodds)
{
  int     r, d;
  int     st;
  int     i, iabs, imod;            /* relative positions in seqX, seqY    */
  int     imodmin;
  int     cur_x, cur_y;       /* nucleotides at those positions      */
  double  fr;
  double  flend, frend;
  double  sc, bestsc; 
  double  foroth;
  int     exact;

  st   = 0; /* this is because we allways grab the sequences from [0,lmax] in a fresh array */

  i    = st    + l - 1;
  iabs = stabs + l - 1;
  imod = (stmod+l-1 > win-1)? stmod+l-1-win : stmod+l-1;
 
  if (imod != iabs%win) Die("bad coor transformation in ViterbiRNADiagScan()");  /* paranoia */

  imodmin = (imod-1 < 0)? imod-1+win : imod-1;

  exact = FALSE;  /* if TRUE due the right thing for the O models. It is toooooo slow! */
  
  if (lmax == 0) 
    return cod->t[TCOBCOE] +
      ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, l, l,
			 cod->COB, dp->cob) +
      ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, l, l,
			 cod->COE, dp->coe);
  
 if (stabs == 0) 
    for (r = 0; r <= imod+1; r++) 
       FillJOTHScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, i, imod, r, cod->COJ, 
		    null, dp->coj, dp->cod->COJ, exact, FALSE);
 else 
     FillJOTHScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, 0, imod, l, cod->COJ,  
		  null, dp->coj, dp->cod->COJ, exact, FALSE);
  
  
  /* state COB
   */
  bestsc = -BIGFLOAT;
  dp->cod->cobmx[imod] = ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, l, l, cod->COB, dp->cob);
  
  /* state C33
   */
  dp->cod->c33mx[imod] = bestC33diagscan(win, i, imod, l, sqinfoX, seqX, sqinfoY, seqY, dp->cod, cod);
  
  /* state Ce
   */
  bestsc = -BIGFLOAT;
  if ((sc = bestCediagscan(win, i, imod, l, sqinfoX, seqX, sqinfoY, seqY, dp, cod, null, ali, logodds)) > bestsc) 
    bestsc = sc;
  dp->cod->cemx[imod] = bestsc;
if(stabs==3&&i==18)printf("::::: %d %d %d %f %f %f\n", 
stabs, i, imod, dp->cod->cobmx[imod], dp->cod->cemx[imod], bestsc);

  /* state COJ
   */
  bestsc = -BIGFLOAT;
  for (d = 0; d < l; d++) 
    {
      if ((sc = dp->cod->cemx[(imod-d<0)?imod-d+win:imod-d] + cod->t[TCeCOJ] + dp->cod->COJ[imod][d]) 
	  > bestsc) bestsc = sc;
      
      if (d > 0 && 
	  (sc = dp->cod->cbmx[(imod-d<0)?imod-d+win:imod-d] + cod->t[TCbCe] + cod->t[TCeCOJ] + dp->cod->COJ[imod][d]) 
	  > bestsc) bestsc = sc;
    }
  dp->cod->cojmx[imod] = bestsc;
  
  /* state Cb
   */
  bestsc = -BIGFLOAT;
  if ((sc = cod->t[TCOBCb] + dp->cod->cobmx[imod]) > bestsc) bestsc = sc;
  if ((sc = cod->t[TCOJCb] + dp->cod->cojmx[imod]) > bestsc) bestsc = sc;
  dp->cod->cbmx[imod] = bestsc;

  /* state COE
   */
  bestsc = -BIGFLOAT;
  if (l == lmax) {

    bestsc = cod->t[TCOBCOE]
      + ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, 0, 0, cod->COB, dp->cob)
      + ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, l, l, cod->COE, dp->coe);

    for (d = 1; d <= l; d++) {
      for (r = 0; r <= l-d; r++) 
	foroth = ViterbiOTHDiagScan(ofp, sqinfoX, seqX+d, sqinfoY, seqY+d, win, stmod+d, r, l-d, cod->COE, dp->coe);
      
      if ((sc = foroth + cod->t[TCeCOE]  + dp->cod->cemx[(stmod+d-1>win-1)? stmod+d-1-win:stmod+d-1])  > bestsc) bestsc = sc;
      if ((sc = foroth + cod->t[TCOBCOE] + dp->cod->cobmx[(stmod+d-1>win-1)? stmod+d-1-win:stmod+d-1]) > bestsc) bestsc = sc;
    }
  }
  
  return bestsc;
}

double 
bestC33diagscan(int win, int i, int imod, int l, SQINFO sqinfoX, int *seqX, SQINFO sqinfoY, int *seqY, 
		struct coddp_s *dp, struct codmodel_s *cod)
{
  int    x1,  x2,  x3;
  int    xx1, xx2, xx3;
  int    y1,  y2,  y3;
  int    yy1, yy2, yy3;
  int    codX;
  int    codY;
  int    stmod;
  int    imod3;
  int    k, kmod;
  double sc, bestsc;
  double hexa, extra_hexamer;

  bestsc = -BIGFLOAT;

  stmod = (imod-l+1 < 0)? imod-l+1+win : imod-l+1;
  imod3 = (imod-3   < 0)? imod-3+win   : imod-3;

  /* TRIPLETS */
  if (i > 1) 
    { 
      x1 = seqX[i-2];
      x2 = seqX[i-1];
      x3 = seqX[i];
      
      y1 = seqY[i-2];
      y2 = seqY[i-1];
      y3 = seqY[i];

      codX = CODON( x1, x2, x3); 
      codY = CODON( y1, y2, y3); 
      
      if ( isitagap(x1) || isitagap(x2) || isitagap(x3) ||
	   isitagap(y1) || isitagap(y2) || isitagap(y3)   )
	{ 
	  return bestsc;
	}
      
    
      if (i > 2 &&
	  (sc = cod->t[TCbC33] + cod->pcodon[codX][codY] + dp->cbmx[imod3])
	  > bestsc) bestsc = sc;
      
      for (k = 2; k < i-2; k++) {
	
	kmod = (stmod+k > win-1)? stmod+k-win : stmod+k;

	xx1 = seqX[k-2];
	xx2 = seqX[k-1];
	xx3 = seqX[k];
	
	yy1 = seqY[k-2];
	yy2 = seqY[k-1];
	yy3 = seqY[k];
	
	hexa = 0.0;
	if ( !isitagap(xx1) && !isitagap(xx2) && !isitagap(xx3) )
	  hexa += EXP2(cod->phexa[CODON(xx1,xx2,xx3)][codX]);
	if ( !isitagap(yy1) && !isitagap(yy2) && !isitagap(yy3) )
	  hexa += EXP2(cod->phexa[CODON(yy1,yy2,yy3)][codY]);
	
	extra_hexamer = (hexa > 0.0)? -1.0 + LOG2(hexa) : 0.0;
	
	if ((sc = 
	     dp->COJ[imod3][i-3-k]   + 
	     cod->t[TC33Ce]          +  
	     cod->t[TCeCOJ]          +  
	     cod->t[TCOJCb]          +  
	     cod->t[TCbC33]          + 
	     cod->pcodon[codX][codY] + 
	     extra_hexamer           +
	     dp->c33mx[kmod]          ) > bestsc) bestsc = sc;
      }
    }
  return bestsc;
}

double 
bestCediagscan(int win, int i, int imod, int l, 
	       SQINFO sqinfoX, int *seqX, SQINFO sqinfoY, int *seqY, 
	       struct coddpscan_s *dp, 
	       struct codmodel_s  *cod, 
	       struct nullmodel_s *null, 
	       struct ali_s       *ali, 
	       int logodds)
{
  int    x1, x2, x3, x4;
  int    y1, y2, y3, y4;
  int    xx1, xx2, xx3;
  int    yy1, yy2, yy3;
  int    tr1, tr2;
  int    codX, cod5X;
  int    codY, cod5Y;
  int    start;
  int    flag;
  int    stmod;
  int    imod3, imod4;
  double add;
  double sc, bestsc;
  double hexa, extra_hexamer;
  
  stmod = (imod-l+1 < 0)? imod-l+1+win : imod-l+1;
  imod3 = (imod-3   < 0)? imod-3+win   : imod-3;
  imod4 = (imod-4   < 0)? imod-4+win   : imod-4;

  /* Cb --> Ce (not considered here but in COJ)
   */
  bestsc = -BIGFLOAT;

  /* TRIPLETS */
  if (i > 1) 
    { 
      x1 = seqX[i-2];
      x2 = seqX[i-1];
      x3 = seqX[i];
      
      y1 = seqY[i-2];
      y2 = seqY[i-1];
      y3 = seqY[i];
      
      codX  = CODON( x1, x2, x3); 
      codY  = CODON( y1, y2, y3); 
      cod5X = CODON5(x1, x2, x3); 
      cod5Y = CODON5(y1, y2, y3);

      xx1 = (i > 4)? seqX[i-5] : 4;
      xx2 = (i > 3)? seqX[i-4] : 4;
      xx3 = (i > 2)? seqX[i-3] : 4;
      
      yy1 = (i > 4)? seqY[i-5] : 4;
      yy2 = (i > 3)? seqY[i-4] : 4;
      yy3 = (i > 2)? seqY[i-3] : 4;

      hexa = 0.0;
      flag = 0;

      if ( !isitagap(x1)  && !isitagap(x2)  && !isitagap(x3)  &&
	   !isitagap(xx1) && !isitagap(xx2) && !isitagap(xx3) ) 
	{
	  hexa += EXP2(cod->phexa[CODON(xx1,xx2,xx3)][codX]);
	  flag++;
	}
      if ( !isitagap(y1)  && !isitagap(y2)  && !isitagap(y3)  &&
	   !isitagap(yy1) && !isitagap(yy2) && !isitagap(yy3) )
	{
	  hexa += EXP2(cod->phexa[CODON(yy1,yy2,yy3)][codY]);
	  flag++;
	}
      
      switch (flag) 
	{
	case 0: extra_hexamer =  0.0;              break;
	case 1: extra_hexamer =        LOG2(hexa); break;
	case 2: extra_hexamer = -1.0 + LOG2(hexa); break;
	}

      if (i == 2) 
	add = cod->t[TCOBCb] + 
	  ViterbiOTHDiagScan(stdout, sqinfoX, seqX, sqinfoY, seqY, win, stmod, 0, 0, cod->COB, dp->cob);
      else 
	add = dp->cod->cbmx[imod3];

      add += extra_hexamer;
 
      if (iscomplete(cod5X)) 
	{
	  if (iscomplete(cod5Y)) {
	    /* Cb --> C33 --> Ce 
	     */
	    tr1 = TCbC33; tr2 = TC33Ce; 
	    if ((sc = cod->t[tr1] + cod->t[tr2] + cod->pcodon[codX][codY])
	      > bestsc) bestsc = sc;
	  }
	  
	  else if (isgap1(cod5Y)) {
	    /* Cb --> C32 --> Ce 
	     */
	    tr1 = TCbC32; tr2 = TC32Ce; 
	    if ((sc = cod->t[tr1] + cod->t[tr2] + 
	      average(1, y2, y3, codX, cod, null, logodds)) 
		> bestsc) bestsc = sc;
	  }
	  
	  else if (isgap2(cod5Y)) {
	    /* Cb --> C32 --> Ce 
	     */
	    tr1 = TCbC32; tr2 = TC32Ce; 
	    if ((sc  = cod->t[tr1] + cod->t[tr2] + 
	      average(2, y1, y3, codX, cod, null, logodds))
		> bestsc) bestsc = sc;
	  }
	  
	  else if (isgap3(cod5Y)) {
	    /* Cb --> C32 --> Ce 
	     */
	    tr1 = TCbC32; tr2 = TC32Ce; 
	    if ((sc = cod->t[tr1] + cod->t[tr2] + 
	      average(3, y1, y2, codX, cod, null, logodds))
		> bestsc) bestsc = sc;
	  }
	  
	  else if (isnocod(cod5Y)) { 
	    /* Cb --> C30 --> Ce 
	     */
	    tr1 = TCbC30; tr2 = TC30Ce; 
	    if ((sc  = cod->t[tr1] + cod->t[tr2] + 
	      average_codon(codX, cod, null, logodds))
		> bestsc) bestsc = sc;
	  }

	}
      else if (iscomplete(cod5Y)) 
	{
	  if (isgap1(cod5X)) {
	    /* Cb --> C23 --> Ce 
	     */
	    tr1 = TCbC23; tr2 = TC23Ce; 
	    if ((sc = cod->t[tr1] + cod->t[tr2] + 
	      average(1, x2, x3, codY, cod, null, logodds)) 
		> bestsc) bestsc = sc;
	  }
	  
	  else if (isgap2(cod5X)) {
	    /* Cb --> C23 --> Ce 
	     */
	    tr1 = TCbC23; tr2 = TC23Ce; 
	    if ((sc = cod->t[tr1] + cod->t[tr2] + 
	      average(2, x1, x3, codY, cod, null, logodds)) 
		> bestsc) bestsc = sc;
	  }
	  
	  else if (isgap3(cod5X)) {
	    /* Cb --> C23 --> Ce 
	     */
	      tr1 = TCbC23; tr2 = TC23Ce; 
	      if ((sc = cod->t[tr1] + cod->t[tr2] + 
		average(3, x1, x2, codY, cod, null, logodds)) 
		> bestsc) bestsc = sc;
	    }

	  else if (isnocod(cod5X)) {
	    /* Cb --> C03 --> Ce 
	     */
	    tr1 = TCbC03; tr2 = TC03Ce; 
	    if ((sc = cod->t[tr1] + cod->t[tr2] + 
	      average_codon(codY, cod, null, logodds)) 
		> bestsc) bestsc = sc;
	  }
	}
    }
  bestsc += add;

  /* QUADRUPLETS */
  if (i > 2) 
    {
      x1 = seqX[i-3];
      x2 = seqX[i-2];
      x3 = seqX[i-1];
      x4 = seqX[i]; /* for codons of four */
      
      y1 = seqY[i-3];
      y2 = seqY[i-2];
      y3 = seqY[i-1];
      y4 = seqY[i]; /* for codons of four */

      codX  = CODON( x1, x2, x3); 
      codY  = CODON( y1, y2, y3); 
      cod5X = CODON5(x1, x2, x3); 
      cod5Y = CODON5(y1, y2, y3);
      
      xx1 = (i > 5)? seqX[i-6] : 4;
      xx2 = (i > 4)? seqX[i-5] : 4;
      xx3 = (i > 3)? seqX[i-4] : 4;
      
      yy1 = (i > 5)? seqY[i-6] : 4;
      yy2 = (i > 4)? seqY[i-5] : 4;
      yy3 = (i > 3)? seqY[i-4] : 4;

      hexa = 0.0;
      flag = 0;

      if ( !isitagap(x1)  && !isitagap(x2)  && !isitagap(x3)  &&
	   !isitagap(xx1) && !isitagap(xx2) && !isitagap(xx3) ) 
	{
	  hexa += EXP2(cod->phexa[CODON(xx1,xx2,xx3)][codX]);
	  flag++;
	}
      if ( !isitagap(y1)  && !isitagap(y2)  && !isitagap(y3)  &&
	   !isitagap(yy1) && !isitagap(yy2) && !isitagap(yy3) )
	{
	  hexa += EXP2(cod->phexa[CODON(yy1,yy2,yy3)][codY]);
	  flag++;
	}
      
      switch (flag) 
	{
	case 0: extra_hexamer = 0.0;               break;
	case 1: extra_hexamer =        LOG2(hexa); break;
	case 2: extra_hexamer = -1.0 + LOG2(hexa); break;
	}
	      
       if (i == 3) 
	add =  cod->t[TCOBCb] +
	  ViterbiOTHDiagScan(stdout, sqinfoX, seqX, sqinfoY, seqY, win, stmod, 0, 0, cod->COB, dp->cob);
      else 
	add = dp->cod->cbmx[imod4];

       add += extra_hexamer;

       if (iscomplete(cod5X)) 
	{
	  if      (iscomplete(cod5Y)) {
	    if (!isitagap(x4) && !isitagap(y4)) {
	      /* Cb --> C44 --> Ce 
	       */
	      tr1 = TCbC44; tr2 = TC44Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + cod->pcodon[codX][codY] + 
		   ((logodds)?0:null->xem[x4]) + ((logodds)?0:null->yem[y4]))
		  > bestsc) bestsc = sc;
	    }
	    else if(!isitagap(x4) && isitagap(y4)) {
	      /* Cb --> C43 --> Ce 
	       */
	      tr1 = TCbC43; tr2 = TC43Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + cod->pcodon[codX][codY] + 
		   ((logodds)?0:null->xem[x4]))
		  > bestsc) bestsc = sc;
	    }
	    else if(isitagap(x4) && !isitagap(y4)) {
	      /* Cb --> C34 --> Ce 
	       */
	      tr1 = TCbC34; tr2 = TC34Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + cod->pcodon[codX][codY] + 
		   ((logodds)?0:null->yem[y4]))
		  > bestsc) bestsc = sc;
	    }
	    else if(isitagap(x4) && isitagap(y4)) {
	      /* Cb --> C33 --> Ce 
	       */
	      tr1 = TCbC33; tr2 = TC33Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + cod->pcodon[codX][codY])
		  > bestsc) bestsc = sc;
	    }
	  }
	  
	  else if (isgap1(cod5Y)) {
	    if ((sc = add +
		 fourcodonX(x1, 1, CODON5(x2, x3, x4), CODON5(y2, y3, y4), cod, null, 
			    logodds, &tr1, &tr2)) 
		> bestsc) bestsc = sc;
	  }
	  else if (isgap2(cod5Y)) {
	    if ((sc = add +
	      fourcodonX(x2, 2, CODON5(x1, x3, x4), CODON5(y1, y3, y4), cod, null, 
			 logodds, &tr1, &tr2)) 
		> bestsc) bestsc = sc;
	  }
	  else if (isgap3(cod5Y)) {
	    if ((sc = add +
	      fourcodonX(x3, 3, CODON5(x1, x2, x4), CODON5(y1, y2, y4), cod, null, 
			 logodds, &tr1, &tr2)) 
		> bestsc) bestsc = sc;
	  }

	  else if (isnocod(cod5Y)) {
	    if      (!isitagap(x4) && !isitagap(y4)) {
	      /* Cb --> C40 --> Ce 
	       */
	      tr1 = TCbC40; tr2 = TC40Ce;
	      if ((sc = add + cod->t[tr1] + average_codon(codX, cod, null, logodds) + 
		   ((logodds)?-null->meta:null->xem[x4]+null->yem[y4]))
		  > bestsc) bestsc = sc;
	    }
	    else if (!isitagap(x4) &&  isitagap(y4)) {
	      /* Cb --> C40 --> Ce 
	       */
	      tr1 = TCbC40; tr2 = TC40Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + average_codon(codX, cod, null, logodds) + 
		   ((logodds)?0:null->xem[x4]))
		  > bestsc) bestsc = sc;
	    }
	    else if ( isitagap(x4) && !isitagap(y4)) {
	      /* Cb --> C30 --> Ce 
	       */
	      tr1 = TCbC30; tr2 = TC30Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + average_codon(codX, cod, null, logodds) + 
		   ((logodds)?-null->meta:null->yem[y4]))
		  > bestsc) bestsc = sc;
	    }
	    else if ( isitagap(x4) &&  isitagap(y4)){
	      /* Cb --> C30 --> Ce 
	       */
	      tr1 = TCbC30; tr2 = TC30Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + average_codon(codX, cod, null, logodds))
		  > bestsc) bestsc = sc;
	    }
	  }

	  else if (!isitagap(x4) && !isitagap(y4)) {
	    if      (miss1(cod5Y) && miss2(cod5Y)) {
	      /* Cb --> C42 --> Ce 
	       */
	      tr1 = TCbC42; tr2 = TC42Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(1, y3, y4, CODON(x2,x3,x4), cod, null, logodds) + ((logodds)?0.:null->xem[x1])) 
		> bestsc) bestsc = sc;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(2, y3, y4, CODON(x1,x3,x4), cod, null, logodds) + ((logodds)?0.:null->xem[x2])) 
		> bestsc) bestsc = sc;
	    }
	    else if (miss1(cod5Y) && miss3(cod5Y)) {
	      /* Cb --> C42 --> Ce 
	       */
	      tr1 = TCbC42; tr2 = TC42Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(1, y2, y4, CODON(x2,x3,x4), cod, null, logodds) + ((logodds)?0.:null->xem[x1])) 
		> bestsc) bestsc = sc;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(3, y2, y4, CODON(x1,x2,x4), cod, null, logodds) + ((logodds)?0.:null->xem[x3])) 
		> bestsc) bestsc = sc;
	    }
	    else if (miss2(cod5Y) && miss3(cod5Y)) {
	      /* Cb --> C42 --> Ce 
	       */
	      tr1 = TCbC42; tr2 = TC42Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(2, y1, y4, CODON(x1,x3,x4), cod, null, logodds) + ((logodds)?0.:null->xem[x2])) 
		> bestsc) bestsc = sc;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(3, y1, y4, CODON(x1,x2,x4), cod, null, logodds) + ((logodds)?0.:null->xem[x3])) 
		> bestsc) bestsc = sc;
	    }
	    else Die("wrong number of gaps in this codon \n");
	  }
	}

      else if (iscomplete(cod5Y))
	{
	  if      (isgap1(cod5X)) {
	    if ((sc = add + 
	      fourcodonY(y1, 1, CODON5(x2, x3, x4), CODON5(y2, y3, y4), cod, null, logodds, &tr1, &tr2))
		> bestsc) bestsc = sc;
	  }
	  else if (isgap2(cod5X)) {
	    if ((sc = add +
	      fourcodonY(y2, 2, CODON5(x1, x3, x4), CODON5(y1, y3, y4), cod, null, logodds, &tr1, &tr2)) 
		> bestsc) bestsc = sc;
	  }
	  else if (isgap3(cod5X)) {
	   if (( sc = add +
	      fourcodonY(y3, 3, CODON5(x1, x2, x4), CODON5(y1, y2, y4), cod, null, logodds, &tr1, &tr2)) 
		> bestsc) bestsc = sc;
	  }

	  else if (isnocod(cod5X)) {
	    if      (!isitagap(x4) && !isitagap(y4)) {
	      /* Cb --> C04 --> Ce 
	       */
	      tr1 = TCbC04; tr2 = TC04Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + average_codon(codY, cod, null, logodds) + 
		   ((logodds)?-null->meta:null->xem[x4]+null->yem[y4]))
		  > bestsc) bestsc = sc;
	    }
	    else if (!isitagap(x4) &&  isitagap(y4)) {
	      /* Cb --> C03 --> Ce 
	       */
	      tr1 = TCbC03; tr2 = TC03Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + average_codon(codY, cod, null, logodds) + 
		   ((logodds)?-null->meta:null->xem[x4]))
		  > bestsc) bestsc = sc;
	    }
	    else if ( isitagap(x4) && !isitagap(y4)) {
	      /* Cb --> C04 --> Ce 
	       */
	      tr1 = TCbC04; tr2 = TC04Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + average_codon(codY, cod, null, logodds) + 
		   ((logodds)?0:null->yem[y4]))
		  > bestsc) bestsc = sc;
	    }
	    else if ( isitagap(x4) &&  isitagap(y4)){
	      /* Cb --> C03 --> Ce 
	       */
	      tr1 = TCbC03; tr2 = TC03Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] + average_codon(codY, cod, null, logodds))
		  > bestsc) bestsc = sc;
	    }
	  }
	  else if (!isitagap(x4) && !isitagap(y4)) {
	    if      (miss1(cod5X) && miss2(cod5X)) {
	      /* Cb --> C24 --> Ce 
	       */
	      tr1 = TCbC24; tr2 = TC24Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(1, x3, x4, CODON(y2,y3,y4), cod, null, logodds) + ((logodds)?0.:null->yem[y1])) 
		> bestsc) bestsc = sc;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(2, x3, x4, CODON(y1,y3,y4), cod, null, logodds) + ((logodds)?0.:null->yem[y2])) 
		> bestsc) bestsc = sc;
       	    }
	    else if (miss1(cod5X) && miss3(cod5X)) {
	      /* Cb --> C24 --> Ce 
	       */
	      tr1 = TCbC24; tr2 = TC24Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(1, x2, x4, CODON(y2,y3,y4), cod, null, logodds) + ((logodds)?0.:null->yem[y1])) 
		> bestsc) bestsc = sc;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(3, x2, x4, CODON(y1,y2,y4), cod, null, logodds) + ((logodds)?0.:null->yem[y3])) 
		> bestsc) bestsc = sc;
	    }
	    else if (miss2(cod5X) && miss3(cod5X)) {
	      /* Cb --> C24 --> Ce 
	       */
	      tr1 = TCbC24; tr2 = TC24Ce;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(2, x1, x4, CODON(y1,y3,y4), cod, null, logodds) + ((logodds)?0.:null->yem[y2])) 
		> bestsc) bestsc = sc;
	      if ((sc = add + cod->t[tr1] + cod->t[tr2] +
		average(3, x1, x4, CODON(y1,y2,y4), cod, null, logodds) + ((logodds)?0.:null->yem[y3])) 
		> bestsc) bestsc = sc;
	    }
	    else Die("wrong number of gaps in this codon");
	  }
	}
    }
  return bestsc;
}

