/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* othdpscan.c
 *
 * ER, Mon Dec  3 14:13:25 CST 2001 [St. Louis]
 * 
 * dynamic programming (viterbi and forward) with the othermodel
 *
 * calculates:
 *                       P(seqX,seqY \pi^* | othemodel)  [viterbi algorithm; \pi^* = best path ]
 *              \sum_\pi P(seqX,seqY \pi   | othemodel)  [forward algorithm ]
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

 /* Function: FillJOTHScan() 
 * Date:     ER, Fri Dec 14 11:28:40 CST 2001 [St. Louis] 
 * 
 * 
 * Returns:  othj[jmod][l]  is filled
 */
void
FillJOTHScan(FILE *ofp, SQINFO sqinfoX, int *seqX, SQINFO sqinfoY, int *seqY, 
	     int win, int j, int jmod, int l, 
	     struct othmodel_s *othmodel, struct nullmodel_s *null, struct othdpd_s *othdp, double  **othj, 
	     int exact, int logodds)
{
  if (exact)
    othj[jmod][l] = ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, j-l, l, l, othmodel, othdp);
  else 
    {
      if (l == 0) 
	othj[jmod][l] = ScoreWithOTH(ofp, seqX, seqY, j, l, othmodel, FALSE);
      
      else if (j == 0 || l == 1)
	othj[jmod][l] = ScoreWithOTH(ofp, seqX, seqY, j, l, othmodel, FALSE) 
	  + ((logodds)? 2.0*null->meta : 0.0);

      else
	othj[jmod][l] = othj[(jmod-1<0)?jmod-1+win:jmod-1][l-1] 
	  + ScoreWithOTHMatrix(seqX, seqY, j, othmodel) 
	  + ((logodds)? 2.0*null->meta : 0.0);
    }
}


/* Function: ViterbiOTHDiagScan() 
 * Date:     ER, Tue Dec  4 09:51:47 CST 2001 [St. Louis] 
 * 
 * Purpose:  Calculates, i \in [start, start+L-1] --- l \in [1, i+1]
 *           mt[i][l] = P(X,Y, begin = start+i-l+1, end = start+i, best align | OTH model) 
 * 
 * Args:     ofp          -- output file 
 *           seqX, seqY   -- equal length sequences, ACGT and gaps (-)  
 *           start        -- start position 
 *           L            -- length of sX,sY  
 *           oth          -- oth_model structure 
 *           dp           -- matrices for dynamic programming
 *           ali          -- array with alignment
 *           traceback    -- if true due traceback
 *           alignment    -- if true print alignment
 * 
 * Returns:  void.
 */
double
ViterbiOTHDiagScan(FILE *ofp, SQINFO sqinfoX, int *seqX, SQINFO sqinfoY, int *seqY, 
		   int win, int stmod, int l, int lmax, struct othmodel_s *oth, struct othdpd_s *othdp)
{
  int     st;
  int     i, imod;            /* relative positions in seqX, seqY    */
  int     imodmin;
  int     d; 
  int     cur_x, cur_y;       /* nucleotides at those positions      */
  double  fr;
  double  flend, frend;
  double  sc, bestsc; 
  
  if (lmax == 0) return oth->t[TFLFR]                             
		   + ScoreWithNullDiag(seqX, seqY, st, l, oth->FLN)
		   + ScoreWithNullDiag(seqX, seqY, st, 0, oth->FRN);
  
  st   = 0; /* this is because we grad the sequences from [0,win] in a fresh array */
  
  i    = st    + l - 1;
  imod = stmod + l - 1;

  imodmin = (imod-1 < 0)? imod-1+win : imod-1;

  flend = ScoreWithNullDiag(seqX, seqY, st, 0, oth->FLN);
  frend = ScoreWithNullDiag(seqX, seqY, st, 0, oth->FRN);
  
  cur_x = seqX[i];
  cur_y = seqY[i];
      
  if (i == st) 
    othdp->flmx[imod] = ScoreWithNullDiag(seqX, seqY, i, l, oth->FLN);
  else 
    othdp->flmx[imod] = othdp->flmx[imodmin] + NullAddOnePair(cur_x, cur_y, oth->FLN);
    
      
  othdp->fjmx[imod][0] = ScoreWithNullDiag(seqX, seqY, i, 0, oth->FJN);
  for (d = 1; d <= l; d++) {
    if (i == st)
      othdp->fjmx[imod][d] = ScoreWithNullDiag(seqX, seqY, i-d+1, d, oth->FJN);
    else
      othdp->fjmx[imod][d] = othdp->fjmx[imodmin][d-1] + NullAddOnePair(cur_x, cur_y, oth->FJN);
  }
  
  /* state B(i) 
   */
  bestsc = oth->t[TFLB] + othdp->flmx[imod];
  for (d = 1; d < l; d++) 
    if ((sc = othdp->emx[(stmod+d-1<0)?stmod+d-1+win:stmod+d-1] + oth->t[TEFJ] + oth->t[TFJB] + 
	 othdp->fjmx[imod][l-d]) > bestsc) bestsc = sc;
  othdp->bmx[imod] = bestsc;
  
  /* state M(i)
   */
  bestsc = -BIGFLOAT;
  if (cur_x < 4 && cur_y < 4) {
    if (i == st) 
      bestsc = oth->t[TBM] + oth->t[TFLB] + flend;
    else {
      if ((sc = oth->t[TBM] + othdp->bmx[imodmin]) > bestsc) bestsc = sc;
      if ((sc = oth->t[TMM] + othdp->mmx[imodmin]) > bestsc) bestsc = sc;
      if ((sc = oth->t[TXM] + othdp->xmx[imodmin]) > bestsc) bestsc = sc;
      if ((sc = oth->t[TYM] + othdp->ymx[imodmin]) > bestsc) bestsc = sc;
    }
    bestsc += oth->mem[idx(cur_x,cur_y)];
  }
  othdp->mmx[imod] = bestsc;
  
  /* state X(i)
   */
  bestsc = -BIGFLOAT;
  if (cur_x < 4 && cur_y == 4) {
    if (i == st) 
      bestsc = oth->t[TBX] + oth->t[TFLB] + flend;
    else {
      if ((sc = oth->t[TBX] + othdp->bmx[imodmin]) > bestsc) bestsc = sc;
      if ((sc = oth->t[TMX] + othdp->mmx[imodmin]) > bestsc) bestsc = sc;
      if ((sc = oth->t[TXX] + othdp->xmx[imodmin]) > bestsc) bestsc = sc;
      if ((sc = oth->t[TYX] + othdp->ymx[imodmin]) > bestsc) bestsc = sc;
    }
    bestsc += oth->xem[cur_x];
  }
  othdp->xmx[imod] = bestsc;
  
  /* state Y(i)
   */
  bestsc = -BIGFLOAT;
  if (cur_x == 4 && cur_y < 4) { 
    if (i == st) 
      bestsc = oth->t[TBY] + oth->t[TFLB] + flend;
    else {
      if ((sc = oth->t[TBY] + othdp->bmx[imodmin]) > bestsc) bestsc = sc;
      if ((sc = oth->t[TMY] + othdp->mmx[imodmin]) > bestsc) bestsc = sc;
      if ((sc = oth->t[TXY] + othdp->xmx[imodmin]) > bestsc) bestsc = sc;
      if ((sc = oth->t[TYY] + othdp->ymx[imodmin]) > bestsc) bestsc = sc;
    }
    bestsc += oth->yem[cur_y];
  }
  othdp->ymx[imod] = bestsc;
  
  /* state E(i)
   */
  bestsc = -BIGFLOAT;
  if ((sc = oth->t[TME] + othdp->mmx[imod]) > bestsc) bestsc = sc;
  if ((sc = oth->t[TXE] + othdp->xmx[imod]) > bestsc) bestsc = sc;
  if ((sc = oth->t[TYE] + othdp->ymx[imod]) > bestsc) bestsc = sc;
  othdp->emx[imod] = bestsc;
  
  /* mt[i][l] = \max_{d=st}^{d=i} E(d)*FRN(d+1 to i)
   *
   * test: mt[i][l] = ViterbiOTHDiag_L(start=iabs-l-1, L=l)
   */
  bestsc = -BIGFLOAT;
  if (l == lmax) {
    for (d = 1; d <= l; d++) {
      fr = ScoreWithNullDiag(seqX, seqY, st+d, l-d, oth->FRN);
      
      if ((sc = othdp->emx[(stmod+d-1<0)? stmod+d-1+win:stmod+d-1]  + oth->t[TEFR]  + fr) > bestsc) bestsc = sc;
      if ((sc = othdp->flmx[(stmod+d-1<0)? stmod+d-1+win:stmod+d-1] + oth->t[TFLFR] + fr) > bestsc) bestsc = sc; 
    }
  }
  
  return bestsc;
}
