/* QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* rnadpscan.c
 *
 * ER, Mon Dec 10 10:14:48 CST 2001 [St. Louis]
 * 
 * dynamic programming (viterbi and forward) with the RNAmodel
 *
 * calculates:
 *                       P(seqX,seqY \pi^* | RNAmodel)  [viterbi algorithm; \pi^* = best path ]
 *              \sum_\pi P(seqX,seqY \pi   | RNAmodel)  [forward algorithm ]
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>


#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

/* Function: AllocDpDiagScanRNA()
 * Date:     ER, Mon Dec 10 10:15:10 CST 2001 [St. Louis]
 *
 * Purpose:  Allocates memory for the dp matrices of the RNA model - scanning diagonal DP
 *
 * Returns:  rnadp are allocated
 */
struct rnadpscan_s *
AllocDpDiagScanRNA(int L)
{
  struct rnadpscan_s *rnadpscan;       /* structure with dp matrices   */

  rnadpscan = (struct rnadpscan_s *) MallocOrDie (sizeof(struct rnadpscan_s));

  rnadpscan->rob = AllocDpDiagOTH(L);
  rnadpscan->roj = AllocDpDiagOTH(L);
  rnadpscan->roe = AllocDpDiagOTH(L);

  rnadpscan->rna = AllocDpDiagRNA(L);
  
  return rnadpscan;
}

/* Function: FreeDpScanRNA()
 * Date:     ER,  Mon Dec 10 10:29:37 CST 2001 [St. Louis]
 *
 * Purpose:  frees memory for the dp matrices of the RNA model
 *
 * Returns:  coddp are freed
 */
void
FreeDpScanRNA(struct rnadpscan_s *rnadpscan)
{
  FreeDpDiagOTH(rnadpscan->rob);
  FreeDpDiagOTH(rnadpscan->roj);
  FreeDpDiagOTH(rnadpscan->roe);

  FreeDpRNA(rnadpscan->rna);

  free(rnadpscan);
}

/* Function: ViterbiRNADiagScan()
 * Date:     ER,  Tue Dec 11 18:01:35 CST 2001[St. Louis]
 *
 * Purpose:  Calculates P(X,Y, best_align | RNA model)
 *
 * Strategy: The algorith is O(L) in time because we respect the  gaps
 *               so there is only one "length".
 *           The algorithm is O(L) in memory, but it can be implemented keeping 
 *               only the previous value.
 *
 * Args:     ofp          -- output file
 *           seqX, seqY   -- equal length sequences, ACGT and gaps (-) 
 *           start        -- start position
 *           d            -- length of sX,sY 
 *           rnamodel     -- rna_model structure
 *
 * Returns:  log likelihood, log P(seqX,seqY,\pi_{blast} | RNAmodel)
 */
double
ViterbiRNADiagScan(FILE *ofp, SQINFO sqinfoX, int *seqX, SQINFO sqinfoY, int *seqY, 
		   int win, int stabs, int stmod, int l, int lmax, 
		   struct model_s *model, struct rnadpscan_s *dp, struct rnascfgscan_s *mx, 
		   struct ali_s *ali, int alignment, int cyk, int logodds, int revstrand, int traceback)
{
  struct rnamtxscan_s *mtx;
  int     d, r;
  int     st;
  int     i, iabs, imod;            /* relative positions in seqX, seqY    */
  int     imodmin;
  int     cur_x, cur_y;             /* nucleotides at those positions     */
  int     exact;
  double  sc, bestsc; 
  double  roe;

  if (revstrand) mtx = mx->inrv;
  else           mtx = mx->in;

  st   = 0;                /* this is because we allways grab the sequences from [0,lmax] in a fresh array */
  i    = st    + l - 1;
  iabs = stabs + l - 1;
  imod = (stmod+l-1 > win-1)? stmod+l-1-win : stmod+l-1;
  if (imod != iabs%win) Die("bad coor transformation in ViterbiRNADiagScan()");  /* paranoia */

  imodmin = (imod-1 < 0)? imod-1+win : imod-1;

  exact = FALSE;  /* if TRUE due the right thing for the O models. It is toooooo slow! */
    
  cur_x = seqX[i];
  cur_y = seqY[i];

  if (lmax == 0) 
    return model->rna->t[TCOBCOE] +
      ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, l, l,
			 model->rna->ROB, dp->rob) +
      ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, l, l,
			 model->rna->ROE, dp->roe);
  
  /* SCFG component
   */
  /* Fill DP-matrices  from [jmod-d, jmod]
   *
   * Correspondences between [stabs,iabs] and   [jmod-d, jmod]
   * are a little weird, because they are filled in different directions,
   * but they can be made work with the following correspondences:
   *
   *  if (stabs == 0) ----->  jabs = iabs  ^  d in [0,jabs]
   *
   *
   *  else            ----->  jabs = iabs  ^  d = l-1
   *
   */
  if (cyk) CYKRNAScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stabs, stmod, l, model->rna, model->null,
		      mtx, logodds, traceback);
  /* else 
     InsideRNAScan(ofp, sqinfoX, seqX, sqinfoY, seqY, start, L, model->rna, model->null,
     dp->oth, mtx, sc, vp, do_nus, logodds); 
  */
  
  if (stabs == 0) 
    for (r = 0; r <= imod+1; r++) 
       FillJOTHScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, i, imod, r, model->rna->ROJ, 
		    model->null, dp->roj, dp->rna->ROJ, exact, FALSE);
 else 
     FillJOTHScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, 0, imod, l, model->rna->ROJ,  
		  model->null, dp->roj, dp->rna->ROJ, exact, FALSE);

   /* state ROB
   */
  bestsc = -BIGFLOAT;
  dp->rna->rbmx[imod] = ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, l, l, model->rna->ROB, dp->rob);
  
  /* state RNA
   */
  bestsc = model->rna->t[TROBRNA] + mtx->wx[imod][l-1] + 
    ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, 0, 0, model->rna->ROB, dp->rob);

  for (d = 1; d < l; d++) {
    if ((sc = model->rna->t[TROBRNA] + dp->rna->rbmx[(imod-d<0)?imod-d+win:imod-d] + mtx->wx[imod][d-1]) > bestsc) bestsc = sc;
    if ((sc = model->rna->t[TROJRNA] + dp->rna->rjmx[(imod-d<0)?imod-d+win:imod-d] + mtx->wx[imod][d-1]) > bestsc) bestsc = sc;
  }
  dp->rna->rrmx[imod] = bestsc;
  
  /* state ROJ
   */
  bestsc = -BIGFLOAT;
  for (d = 0; d < l; d++) 
    if ((sc = model->rna->t[TRNAROJ] + dp->rna->rrmx[(imod-d<0)?imod-d+win:imod-d] + dp->rna->ROJ[imod][d]) > bestsc) bestsc = sc;
  dp->rna->rjmx[imod] = bestsc;

  /* state ROE
   */
  bestsc = -BIGFLOAT;
  if (l == lmax) {
    bestsc = model->rna->t[TROBROE] + 
      ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, 0, 0, model->rna->ROB, dp->rob) +
      ViterbiOTHDiagScan(ofp, sqinfoX, seqX, sqinfoY, seqY, win, stmod, l, l, model->rna->ROE, dp->roe);
 
    for (d = 1; d < l; d++) {
      for (r = 0; r <= l-d; r++) 
	roe = ViterbiOTHDiagScan(ofp, sqinfoX, seqX+d, sqinfoY, seqY+d, win, stmod+d, r, l-d, model->rna->ROE, dp->roe);
      
      if((sc = model->rna->t[TRNAROE] + dp->rna->rrmx[(stmod+d-1>win-1)? stmod+d-1-win:stmod+d-1] + roe) > bestsc) bestsc = sc;
      if((sc = model->rna->t[TROBROE] + dp->rna->rbmx[(stmod+d-1>win-1)? stmod+d-1-win:stmod+d-1] + roe) > bestsc) bestsc = sc;
    }
  }
  
  return bestsc;
}
