/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* rnascfgscan.c
 *
 * ER, Tue Dec 11 13:45:57 CST 2001 [St. Louis]
 * 
 * dynamic programming (cyk and inside) with the rnamodel
 *
 * calculates:
 *                       P(seqX,seqY \pi^* | rnamodel)  [CYK; \pi^* = best path ]
 *              \sum_\pi P(seqX,seqY \pi   | rnamodel)  [fInside algorithm ]
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

static void bestvpscan(int *sX, int *sY, struct pi2model_s *pi2, int win, int j, int jmod, int d, double *vp, double **vx);

/* Function: AllocRNAMtxScan()
 *
 * Date:     ER,  Wed Dec 12 12:42:54 CST 2001 [St. Louis]
 *
 * Purpose:  allocate dp matrices for the RNA model - scanning version
 * Args:     L    --  length of the sequences
 *           vx   --  dp matrix (L x L)
 *           wx   --  dp matrix (L x L)
 *           wbx  --  dp matrix (L x L)
 *
 * Purpose:  Allocates memory for the dp matrices of the RNA model
 *
 * Returns:  vx, wx, wbx are allocated.
 *
 */
struct rnamtxscan_s *
AllocRNAMtxScan(int L)
{
  struct rnamtxscan_s *mtx;       /* structure with dp matrices   */
  int                  j;

  mtx = (struct rnamtxscan_s *) MallocOrDie (sizeof(struct rnamtxscan_s));

  mtx->sc   = (double  *) MallocOrDie(sizeof(double  ) * (10*L+2));
  mtx->vp   = (double  *) MallocOrDie(sizeof(double  ) * L);
 
  mtx->rnaj = (double **) MallocOrDie(sizeof(double *) * L);
  mtx->vx   = (double **) MallocOrDie(sizeof(double *) * L);
  mtx->wx   = (double **) MallocOrDie(sizeof(double *) * L);
  mtx->wbx  = (double **) MallocOrDie(sizeof(double *) * L);

  mtx->rnaj[0] = (double *) MallocOrDie(sizeof(double) * L * (L+1) / 2);
  mtx->vx[0]   = (double *) MallocOrDie(sizeof(double) * L * (L+1) / 2);
  mtx->wx[0]   = (double *) MallocOrDie(sizeof(double) * L * (L+1) / 2);
  mtx->wbx[0]  = (double *) MallocOrDie(sizeof(double) * L * (L+1) / 2);

  mtx->othj = AllocDpDiagOTH(L);

  for (j = 1; j < L; j++) {
    mtx->rnaj[j] = mtx->rnaj[0] + j*(j+1)/2;
    mtx->vx[j]   = mtx->vx[0]   + j*(j+1)/2;
    mtx->wx[j]   = mtx->wx[0]   + j*(j+1)/2;
    mtx->wbx[j]  = mtx->wbx[0]  + j*(j+1)/2;
  }

  /* Initialize */
  PatternRNAMtxScan(L, mtx);

  return mtx;
}

/* Function: AllocScfgRNA()
 *
 * Date:     ER,  Wed Dec 12 12:49:07 CST 2001[St. Louis]
 *
 * Purpose:  allocate matrices for the RNA model -- scanning version
 *
 * Args:     L    --  length of the sequences
 *           sc   --  dp matrix (4L+2)

 *           vp   --  dp matrix (L)
 *           vx   --  dp matrix (L x L)
 *           wx   --  dp matrix (L x L)
 *           wbx  --  dp matrix (L x L)
 *           rnaj --  dp matrix (L x L)
 * Purpose:  Allocates memory for the dp matrices of the RNA model
 *
 * Returns:  sc, vp, vx, wx, wbx, rnaj are allocated.
 *
 */
struct rnascfgscan_s *
AllocScfgScanRNA(int L)
{
  struct rnascfgscan_s *mx;       /* structure with dp matrices   */
  int               j;

  mx = (struct rnascfgscan_s *) MallocOrDie (sizeof(struct rnascfgscan_s));

  /* Allocate a 0..L-1 square matrix.
   */
  mx->in   = AllocRNAMtxScan(L);
  mx->inrv = AllocRNAMtxScan(L);

  return mx;
}

/* Function: BestV()
 * Date:     ER, Sat Jan  1 10:06:54 CST 2000 [Panticosa]
 *
 * Purpose:  fill matrix V using the CYK algorithm
 *
 * Args:     L    --
 *           vx   --
 *           wx   --
 *           wbx  --
 *           j,d   - coordinates of the matrix element
 *
 * Returns:  void
 *           matrix vx is filled at position [j][d].
 *
 */
void
BestVScan(int *sX, int *sY, struct pi2model_s *pi2, int win, int j, int jmod, int d, struct rnamtxscan_s *mtx)
{
  int     x,y;
  int     i, imod;
  int     jmodmin;
  int     xi,xj,yi,yj;
  int     xk,xl,yk,yl;
  int     mid;
  int     k,l;
  int     lmod;
  int     mid1,mid2;
  int     nt;
  double  scnt;
  double  sc, bestsc;
  
  i = j - d;

  imod    = (jmod-d < 0)? jmod-d+win  : jmod-d;
  jmodmin = (jmod-1 < 0)? jmod-1+win  : jmod-1;

  xi = sX[i];
  yi = sY[i];
  xj = sX[j];
  yj = sY[j];
  
 /* Initialize diagonal 
   */
  if (d < 4) { mtx->vx[jmod][d] = -BIGFLOAT; return; }
  
  bestsc = -BIGFLOAT;
  
  /* Main recursion
   */

  /* hairpin loop
   */
  /* this is the way to calculate IS1s using the pi2->is1->tn[d]
     
     if (d < MAXRNALOOP) {
     scnt = 0.;
     for (nt = 1; nt < d; nt++) {
     x = sX[i+nt];
     y = sY[i+nt];
     scnt += pi2->is1->ps[idx5(x,y)];
     }
     if ((sc = scnt + pi2->v->t1 + pi2->is1->tn[d]) > bestsc) bestsc = sc;
     }
  */
  
  /* this is the way to calculate IS1s using an OTH Model */
  if (d < MAXRNALOOP)  
    if ((sc = pi2->v->t1 
	 + pi2->is1->tn[d] 
	 + mtx->rnaj[jmodmin][d-1]) > bestsc)  
      bestsc = sc; 

  /* stem loops 
   */
  xk = sX[i+1];
  yk = sY[i+1];
  xl = sX[j-1];
  yl = sY[j-1];
  
  /* this is the way to calculate IS2s using the pi2->is2->tn[d]
   */
  if (d > 1)  
    if ((sc = mtx->vx[jmodmin][d-2] 
	 + pi2->v->t2s 
	 + pi2->is2->tn[2] 
	 + pi2->v->pp[idx5(xk,yk)][idx5(xl,yl)]) > bestsc) bestsc = sc;
  
  /* this is the way to calculate IS2s using an OTH Model
   *
   * if (d > 1) if ((sc = vx[j-1][d-2] + pi2->v->t2s +
   *               pi2->v->pp[idx5(xk,yk)][idx5(xl,yl)]) > bestsc) bestsc = sc;
   */
  
  /* bulges L */
  for (mid = 2; mid < d; mid++)    
    if ((mid+1) < MAXRNALOOP) {
      xk = sX[i+mid];
      yk = sY[i+mid];
      xl = sX[j-1];
      yl = sY[j-1];
      
      /* this is the way to calculate IS2s using the pi2->is2->tn[d]
       */
      scnt = 0.;
      for (nt = 1; nt < mid; nt++) {
	x = sX[i+nt];
	y = sY[i+nt];
	scnt += pi2->is2->ps[idx5(x,y)];
      }
      if ((sc = mtx->vx[jmodmin][d-mid-1] 
	   + scnt 
	   + pi2->v->t2b 
	   + pi2->is2->tn[mid+1] 
	   + pi2->v->pp[idx5(xk,yk)][idx5(xl,yl)]) > bestsc) bestsc = sc;
      
       /* this is the way to calculate IS2s using an OTH Model
       *
       * if ((sc = vx[j-1][d-mid-1] + pi2->v->t2b + 
       *   rnaj[i+mid-1][mid-2] +
       *   pi2->v->pp[idx5(xk,yk)][idx5(xl,yl)]) > bestsc) bestsc = sc;
       */
    }
  
  /* bulges R */
  for (mid = 2; mid < d; mid++)    
    if ((mid+1) < MAXRNALOOP) {
      xk = sX[i+1];
      yk = sY[i+1];
      xl = sX[j-mid];
      yl = sY[j-mid];
      
      /* this is the way to calculate IS2s using the pi2->is2->tn[d]
       */
      scnt = 0.;
      for (nt = 1; nt < mid; nt++) {
	x = sX[j-nt];
	y = sY[j-nt];
	scnt += pi2->is2->ps[idx5(x,y)];
      }
      
      if ((sc = mtx->vx[(jmod-mid<0)?jmod-mid+win:jmod-mid][d-mid-1] 
	   + scnt 
	   + pi2->v->t2b 
	   + pi2->is2->tn[mid+1] 
	   + pi2->v->pp[idx5(xk,yk)][idx5(xl,yl)]) > bestsc) bestsc = sc;
      
      /* this is the way to calculate IS2s using an OTH Model
       *
       * if ((sc = vx[j-mid][d-mid-1] + pi2->v->t2b +
       *   rnaj[j-1][mid-2] +
       *   pi2->v->pp[idx5(xk,yk)][idx5(xl,yl)]) > bestsc) bestsc = sc;
       */
      
    }
  
  /* internal loops */

#ifdef L4CYKIL
  /* L^4  calculation of internal loops */
  for (k = i+2; k < j-1; k++)  
    for (l = k; l < j-1; l++) { 
      
      mid1 = k - i; 
      mid2 = j - l; 
      
      lmod = (imod+l>win-1)? imod+l-win : imod+l;

      if ((mid1+mid2+d) < MAXRNALOOP) { 
	scnt = 0.; 
	for (nt = 1; nt < mid1; nt++) { 
	  x = sX[i+nt]; 
	  y = sY[i+nt]; 
	  scnt += pi2->is2->ps[idx5(x,y)]; 
	} 
	for (nt = 1; nt < mid2; nt++) { 
	  x = sX[j-nt]; 
	  y = sY[j-nt]; 
	  scnt += pi2->is2->ps[idx5(x,y)]; 
	} 
	
	if ((sc = mtx->vx[lmod][d-mid1-mid2] 
	     + pi2->v->t2i 
	     + scnt 
	     + pi2->is2->tn[mid1+mid2]  
	     + pi2->v->pp[idx5(sX[k],sY[k])][idx5(sX[l],sY[l])]) > bestsc) bestsc = sc; 
      } 
    } 
#else
  /* L^3 calculation of internal loops */
  bestvpscan(sX, sY, pi2, win, j, jmod, d, mtx->vp, mtx->vx);
  
  if (d < MAXRNALOOP) 
    for (mid = 0; mid <= (d-4); mid++) {
      if ((sc = mtx->vp[mid] + pi2->is2->tn[d-mid]) > bestsc) bestsc = sc;
    }
#endif

  /* multiloops */ 
  for (mid = 3; mid < d-3; mid++) {
    if ((sc = pi2->v->t3 
	 + mtx->wbx[(jmod-d+1+mid<0)?jmod-d+1+mid+win:jmod-d+1+mid][mid] 
	 + mtx->wbx[jmodmin][d-mid-3]) > bestsc) bestsc = sc;
  }

   				/* summation */
  mtx->vx[jmod][d] = bestsc;
}


/* Function: BestWScan()
 * Date:     ER, Wed Dec 12 16:05:14 CST 2001 [St. Louis]
 *
 * Purpose:  fill matrix W using the CYK algorithm -- scanning version
 *
 * Args:     L    --
 *           vx   --
 *           wx   --
 *           wbx  --
 *           j,d   - coordinates of the matrix element
 *
 * Returns:  void
 *           matrix wx is filled at position [j][d].
 *
 */
void
BestWScan(int *sX, int *sY, struct pi2model_s *pi2, int win, int j, int jmod, int d, struct rnamtxscan_s *mtx)
{
  int     st;
  int     xi,xj,yi,yj;
  int     mid;
  int     i, imod;
  int     jmodmin;
  double  sc, bestsc;

  i = j - d;
  imod    = (jmod-d < 0)? jmod-d+win  : jmod-d;
  jmodmin = (jmod-1 < 0)? jmod-1+win  : jmod-1;

  xi = sX[i];
  yi = sY[i];
  xj = sX[j];
  yj = sY[j];
  
  bestsc = -BIGFLOAT;

  /* Initialize diagonal 
   */
  if (d == 0) { mtx->wx[jmod][0] = pi2->w->tl + pi2->w->pl[idx5(xi,yi)]; return; }

  /* Main recursion
   */

                                /* i left; connect to i+1, j; emit xi,yi */
  if ((sc = mtx->wx[jmod][d-1] + pi2->w->tl + pi2->w->pl[idx5(xi,yi)]) > bestsc) bestsc = sc;
  
				/* j right; connect to i, j-1; emit xj,yj */
  if ((sc = mtx->wx[jmodmin][d-1] + pi2->w->tr + pi2->w->pr[idx5(xj,yj)]) > bestsc) bestsc = sc;
  
				/* i,j pair; connect to i+1,j-1; emit xy */
  if ((sc = mtx->vx[jmod][d] + pi2->w->tv + pi2->w->pp[idx5(xi,yi)][idx5(xj,yj)]) > bestsc) bestsc = sc;
  
				/* bifurcations */
  for (mid = 0; mid < d; mid++)
    if ((sc = mtx->wx[jmod][mid] 
	 + mtx->wx[(jmod-mid-1<0)?jmod-mid-1+win:jmod-mid-1][d-mid-1] 
	 + pi2->w->tw) > bestsc) bestsc = sc;
  
				/* choose best */
  mtx->wx[jmod][d] = bestsc;
}

/* Function: BestWB()
 * Date:     ER, Sat Jan  1 09:59:44 CST 2000 [Zaragoza]
 *
 * Purpose:  fill matrix WB using the CYK algorithm
 *
 * Args:     L    --
 *           vx   --
 *           wx   --
 *           wbx  --
 *           j,d   - coordinates of the matrix element
 *
 * Returns:  void
 *           matrix wbx is filled at position [j][b]
 *
 */
void
BestWBScan(int *sX, int *sY, struct pi2model_s *pi2, int win, int j, int jmod, int d, struct rnamtxscan_s *mtx)
{
  int     xi,xj,yi,yj;
  int     mid;
  int     i, imod;
  int     jmodmin;
  double  sc, bestsc;
  
  i = j - d;

  imod    = (jmod-d < 0)? jmod-d+win  : jmod-d;
  jmodmin = (jmod-1 < 0)? jmod-1+win  : jmod-1;

  xi = sX[i];
  yi = sY[i];
  xj = sX[j];
  yj = sY[j];
  
  bestsc = -BIGFLOAT;

  /* wbx has to have structure inside
   */
  if (d == 0) { mtx->wbx[jmod][0] = -BIGFLOAT; return; }

  /* Main recursion
   */
                                /* i left; connect to i+1, j; emit xi,yi */
  if ((sc = mtx->wbx[jmod][d-1] + pi2->wb->tl + pi2->wb->pl[idx5(xi,yi)]) > bestsc) bestsc = sc;

				/* j right; connect to i, j-1; emit xj,yj */
  if ((sc = mtx->wbx[jmodmin][d-1] + pi2->wb->tr + pi2->wb->pr[idx5(xj,yj)]) > bestsc) bestsc = sc;

				/* i,j pair; connect to i+1,j-1; emit xy */
  if ((sc = mtx->vx[jmod][d] 
       + pi2->wb->tv 
       + pi2->wb->pp[idx5(xi,yi)][idx5(xj,yj)]) > bestsc) bestsc = sc;

				/* bifurcations */
  for (mid = 0; mid < d; mid++)
    if ((sc = mtx->wbx[jmod][mid] 
	 + mtx->wbx[(jmod-mid-1<0)?jmod-mid-1+win:jmod-mid-1][d-mid-1] 
	 + pi2->wb->tw) > bestsc) bestsc = sc;

				/* pick the best */
  mtx->wbx[jmod][d] = bestsc;
}


/* Function: CYKRNAScan()
 * Date:     ER, Wed Dec 12 13:21:56 CST 2001 [St. Louis]
 *
 * Purpose:  Score a gapped sequence alignment with RNA model. -- scanning version
 *           Sums over all possible structures.
 *
 */
void
CYKRNAScan(FILE *ofp, SQINFO sqinfoX, int *sX, SQINFO sqinfoY, int *sY, 
	   int win, int stabs, int stmod, int l,
	   struct rnamodel_s *rna, struct nullmodel_s *null, 
	   struct rnamtxscan_s *mtx, int logodds, int traceback)
{
  int i, iabs, imod;
  int j, jabs, jmod;
  int d, r;
  int exact;
  
  exact = FALSE;


  /* ORIGINAL set of coordinates: {stabs, l}
   *
   *          used to fill the HMM-part of the different models.
   *
   *          region ---  [stabs, iabs]     with iabs = stabs + l - 1
   *
   */
  i = l - 1;
  iabs = stabs + i;
  imod = (stmod+i > win-1)? stmod+i-win : stmod+i;
 
  if (imod != iabs%win) Die("bad coor transformation in CYKRNAScan()");  /* paranoia */
  
  /* Fill DP-matrices  from [jabs-d, jabs]
   *
   *   Correspondence between  [stabs,iabs]  and   [jabs-d, jabs]
   *   are a little weird, because they are filled in different directions,
   *   but they can be made work within the same loop 
   *   with the following correspondences:
   *

   *
   *         if (stabs == 0)  ----->  jabs = iabs         ^  d in [0,jmod]
   *
   *         else             ----->  jabs = stabs+win-1  ^  d = l-1
   *
   */

 if (stabs == 0) 
   {
     j    = i;
     jabs = iabs;
     jmod = imod;
     
     /* here I calculate mtx->rnaj[jmod][d+1] for hairpin loops */
     for (r = 0; r <= jmod+1; r++) 
       FillJOTHScan(ofp, sqinfoX, sX, sqinfoY, sY, win, j, jmod, r, rna->pi2->Rloop, 
		    null, mtx->othj, mtx->rnaj, exact, logodds);
     
     for (d = 0; d <= jmod; d++) {
       BestVScan (sX, sY, rna->pi2, win, j, jmod, d, mtx);
       BestWBScan(sX, sY, rna->pi2, win, j, jmod, d, mtx);
       BestWScan (sX, sY, rna->pi2, win, j, jmod, d, mtx);
     }
   }
 else 
   {
     d    = l - 1;
     j    = win - 1;
     jabs = stabs + j;
     jmod = jabs % win;
     
     if (jmod != jabs%win) Die("bad coor transformation in CYKRNAScan()");  /* paranoia */
   
    /* here I calculate mtx->rnaj[jmod][l] for hairpin loops */
     FillJOTHScan(ofp, sqinfoX, sX, sqinfoY, sY, win, 0, stmod%win, l, rna->pi2->Rloop, 
		  null, mtx->othj, mtx->rnaj, exact, logodds);
     
     BestVScan (sX, sY, rna->pi2, win, j, jmod, d, mtx);
     BestWBScan(sX, sY, rna->pi2, win, j, jmod, d, mtx);
     BestWScan (sX, sY, rna->pi2, win, j, jmod, d, mtx);
   }
 
}

void
FreeRNAMtxScan(struct rnamtxscan_s *mtx)
{
  free(mtx->sc);
  free(mtx->vp);

  free(mtx->rnaj[0]);
  free(mtx->vx[0]);
  free(mtx->wx[0]);
  free(mtx->wbx[0]);

  free(mtx->rnaj);
  free(mtx->vx);
  free(mtx->wx);
  free(mtx->wbx);

  FreeDpDiagOTH(mtx->othj);

  free(mtx);
}

void
FreeScfgScanRNA(struct rnascfgscan_s *mx)
{
  FreeRNAMtxScan(mx->in);
  FreeRNAMtxScan(mx->inrv);
  
  free(mx);
}

void
PatternRNAMtxScan(int L, struct rnamtxscan_s *mtx)
{
  int j, d;

  PatternVec(L, mtx->sc);
  PatternVec(L, mtx->vp);
  
  PatternDpDiagOTH(L, mtx->othj);
  
  for (j = 0; j < L; j++) 
    for (d = 0; d <= j; d++) {
      mtx->rnaj[j][d] = -BIGFLOAT;
      mtx->vx[j][d]   = -BIGFLOAT;
      mtx->wx[j][d]   = -BIGFLOAT;
      mtx->wbx[j][d]  = -BIGFLOAT;
    }
}

void
PatternScfgScanRNA(int L, struct rnascfgscan_s *mx)
{
  PatternRNAMtxScan(L, mx->in);
  PatternRNAMtxScan(L, mx->inrv); 
}

/* Function: bestvpscan() 
 * 
 * Date:     ER,  Wed Dec 12 16:41:48 CST 2001  [STL]
 *
 * Purpose: calculate array vp for internal loops for the CYK algorithm
 *           
 * Arguments: sX    - sequenceX (0..len-1) 
 *                    representing array indices of symbols
 *            sY    - sequenceX (0..len-1) 
 *                    representing array indices of symbols
 *            cfg   - context-free grammar state transitions, float log2 form
 *            ntc   - context-free grammar nucleotide composition, float log2 form
 *             vp   - DP matrix, already alloc'ed 
 *             vx   - DP matrix, already alloc'ed 
 *            j,d   - coordinates of the matrix element
 *
 *            
 * Return:    (void)           
 *            vp is filled.
 */       
static void 
bestvpscan(int *sX, int *sY, struct pi2model_s *pi2, int win, int j, int jmod, int d, double *vp, double **vx)
{
  int    mid;  
  int    nt;
  double scntc;
  double sc;
  
  /* initialize vp for mid = d-4 (first possible internal loop)
   */
  if (d < 4) 
    { 
      vp[d] = -BIGFLOAT; 
      return; 
    }
  else if (d > 3) 
    {
      vp[d-4] = vx[(jmod-2<0)?jmod-2+win:jmod-2][d-4] 
	+ pi2->v->t2i 
	+ pi2->is2->ps[idx5(sX[j-1],  sY[j-1])]   
	+ pi2->is2->ps[idx5(sX[j-d+1],sY[j-d+1])] 
	+ pi2->v  ->pp[idx5(sX[j-d+2],sY[j-d+2])][idx5(sX[j-2],sY[j-2])];
    }
  
  for (mid = 0; mid < (d-4); mid++) {
    
    if (vp[mid] > -BIGFLOAT) 
      {
	/* calculate the probability of the "new element" 
	 *     sc+k 
	 *           (sc = )
	 *           (k  = pi2->is2->ps[idx5(sX[j-d+1],sY[j-d+1])])
	 *
	 * then pick the best:
	 *
	 *    vp = best[(vp+k) , (sc+k)] 
	 *       = k + best(vp,sc)
	 */
	scntc = 0.;
	for (nt = 1; nt < (d-mid-2); nt++) 
	  scntc += pi2->is2->ps[idx5(sX[j-nt],sY[j-nt])];
	
	if ((sc = vx[(jmod-d+mid+2<0)?jmod-d+mid+2+win:jmod-d+mid+2][mid] 
	     + scntc 
	     + pi2->v->t2i 
	     + pi2->v->pp[idx5(sX[j-d+2],sY[j-d+2])][idx5(sX[j-d+mid+2],sY[j-d+mid+2])]) > vp[mid]) vp[mid] = sc;
	
	vp[mid] +=  pi2->is2->ps[idx5(sX[j-d+1],sY[j-d+1])];
      }
  } 
}

