/************************************************************
 * QRNA - Comparative analysis of biological sequences 
 *        with pair hidden Markov models and pair stochastic context-free grammars
 * Copyright (C) 2000-Howard Hughes Medical Institute
 * All Rights Reserved
 * 
 *     This source code is distributed under the terms of the
 *     GNU General Public License. See the files COPYING and LICENSE
 *     for details.
 ************************************************************/

/* scorewithmodelscan.c
 *
 * ER, Mon Dec  3 11:55:47 CST 2001 [St. Louis]
 * 
 * 
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>

#include "funcs.h"
#include "globals.h"
#include "squid.h"
#include "structs.h"

#ifdef MEMDEBUG
#include "dbmalloc.h"
#endif


static void score_with_null_scan(FILE *ofp, int *isegX, int *isegY, int len, struct nullmodel_s *null, 
				 double nullsc, double nullrevsc, int verbose);

static void score_viterbi_diag_scan(FILE *ofp, 
				    SQINFO sqinfoX, int *isegX,
				    SQINFO sqinfoY, int *isegY,
				    int win, int st, int stmod, int l, int lmax,
				    struct model_s       *model, 
				    struct dpdscan_s     *dpdscan, 		     		    
				    struct rnascfgscan_s *mx, 
				    struct sc3_s         *sc, 
				    struct ali_s         *ali, 
				    struct scansc_s      *scansc,
				    int alignment, int cyk, int logodds, int rnass, int revstrand, int traceback);

/* Function: PrintScoresScan()
 * Date:     ER, Sun Dec  9 14:49:40 CST 2001 [St. Louis]
 *
 * Purpose:  print scores - scan version
 *
 * Args:     ofp, sqinfoX, sqinfoY, nullsc, pamsc, rnasc
 *
 * Returns:  void.
 */
void
PosteriorScoresScan(FILE *ofp, struct scansc_s *scansc, int stabs, int lmax)
{
  char    *win;
  double   othsc,  codsc,  rnasc;
  double   post_o, post_c, post_r;  /* posterior probs in bits */

  othsc = LOG2(EXP2(scansc->oth[stabs]) + EXP2(scansc->othrev[stabs])) - 1.0;
  codsc = LOG2(EXP2(scansc->cod[stabs]) + EXP2(scansc->codrev[stabs])) - 1.0;
  rnasc = LOG2(EXP2(scansc->rna[stabs]) + EXP2(scansc->rnarev[stabs])) - 1.0;

  if      (othsc >= codsc && othsc >= rnasc) win = "OTH";
  else if (codsc >= othsc && codsc >= rnasc) win = "COD";
  else if (rnasc >= othsc && rnasc >= codsc) win = "RNA";
  else Die ("who is the winner?");
  
  fprintf(ofp, "[%d-%d] **", stabs, stabs+lmax-1);
  
  post_o = PosteriorLog2(othsc, codsc, rnasc);
  post_c = PosteriorLog2(codsc, othsc, rnasc);
  post_r = PosteriorLog2(rnasc, othsc, codsc);

  fprintf(ofp, " OTH = %4.3f(%4.3f)  COD = %4.3f(%4.3f)  RNA = %4.3f(%4.3f) ** %s\n\n", 
	  othsc, post_o-post_o, codsc, post_c-post_o, rnasc, post_r-post_o, win);
}

/* Function: ScoreWithModelsScan()
 * Date:     ER, Sat Feb 26 10:39:26 CST 2000 [St. Louis]
 *
 * Purpose:  decide which scoring method to use. Called by main().
 *
 * Args:     
 *
 * Returns:  void.
 */
void
ScoreWithModelsScan(FILE *ofp, 
		    SQINFO sqinfoX, int *isegX,
		    SQINFO sqinfoY, int *isegY,
		    char *gss,
		    int win, int stabs, int stmod, int l, int lmax, 
		    struct model_s       *model, 
		    struct dpdscan_s     *dpdscan, 		   
		    struct rnascfgscan_s *mx, 
		    struct scores_s      *sc, 
		    struct ali_s         *ali, 
		    struct scansc_s      *scansc,
		    int alignment, int cyk, int logodds, int rnass, int revstrand, int shuffle,
		    int traceback, int verbose)
{
  /* Calculate the null-model scores
   */
  score_with_null_scan(ofp, isegX, isegY, l, model->null, sc->null, sc->nullrev, verbose);

  /* VITERBI (diag local aligmnent) 
   */
  score_viterbi_diag_scan(ofp, sqinfoX, isegX, sqinfoY, isegY, win, stabs, stmod, l, lmax, model, dpdscan, 
			  mx, sc->vidiag, ali, scansc, alignment, cyk, logodds, rnass, revstrand, traceback);
  
}

void
score_with_null_scan(FILE *ofp, int *isegX, int *isegY, int len, struct nullmodel_s *null, 
		     double nullsc, double nullrevsc, int verbose)
{
  nullsc    = ScoreWithNullDiag(isegX, isegY, 0, len, null);
  nullrevsc = ScoreWithNullRevDiag(isegX, isegY, 0, len, null);
  if (verbose) PrintNull(ofp, null, nullsc, nullrevsc);
}

void
score_viterbi_diag_scan(FILE *ofp, 
			SQINFO sqinfoX, int *isegX, 
			SQINFO sqinfoY, int *isegY, 
			int win, int stabs, int stmod, int l, int lmax,
			struct model_s       *model, 
			struct dpdscan_s     *dpdscan, 		     		    
			struct rnascfgscan_s *mx, 
			struct sc3_s         *sc, 
			struct ali_s         *ali, 
			struct scansc_s      *scansc,
			int alignment, int cyk, int logodds, int rnass, int revstrand, int traceback)
{
  double othsc;
  double codsc;
  double rnasc;

  othsc = ViterbiOTHDiagScan(ofp, sqinfoX, isegX, sqinfoY, isegY, win, stmod, l, lmax,
			     model->oth, dpdscan->oth);

  codsc = ViterbiCODDiagScan(ofp, sqinfoX, isegX, sqinfoY, isegY, win, stabs, stmod, l, lmax,
			     model->cod, dpdscan->codscan, model->null, ali, alignment, logodds); 
  
   rnasc = ViterbiRNADiagScan(ofp, sqinfoX, isegX, sqinfoY, isegY, win, stabs, stmod, l, lmax,
			     model, dpdscan->rnascan, mx, ali, alignment, cyk, logodds, revstrand, traceback);
  
  if (l == lmax) printf("%d> %f %f %f\n",lmax, othsc,codsc,rnasc);

  /* store scores
   */
  if (l == lmax) 
    if (revstrand) { 
      scansc->othrev[stabs] = othsc;
      scansc->codrev[stabs] = codsc;
      scansc->rnarev[stabs] = rnasc;
      
    }
    else { 
      scansc->oth[stabs] = othsc;
      scansc->cod[stabs] = codsc;
      scansc->rna[stabs] = rnasc;
    }
}
